from pathlib import Path

from l10n._extractor import extract_messages


ROOT = Path(__file__).parent.parent


def test_extract_messages__example():
    path = ROOT / 'example-project'
    messages = list(extract_messages(path))
    assert [m.text for m in messages] == [
        'Hello, {user_name}!',
    ]



from argparse import ArgumentParser

from l10n import Locales


locales = Locales()


def say_hello(name: str, lang: str = 'en') -> None:
    loc = locales[lang]
    msg = loc.get('Hello, {user_name}!').format(user_name=name)
    print(msg)


def main():
    parser = ArgumentParser()
    parser.add_argument('--name', default='World')
    parser.add_argument('--lang', default='en')
    args = parser.parse_args()
    say_hello(name=args.name, lang=args.lang)



from pathlib import Path

import pytest


@pytest.fixture
def project_root(tmp_path: Path) -> Path:
    root = tmp_path / 'project-test'
    root.mkdir()
    (root / 'pyproject.toml').write_text('')
    (root / 'project_test').mkdir()
    (root / 'project_test' / '__init__.py').write_text('')
    return root


@pytest.fixture
def source_path(project_root: Path) -> Path:
    path = (project_root / 'project_test' / 'core.py')
    path.write_text('')
    return path



from __future__ import annotations

import datetime
import gettext
import locale
import threading
from contextlib import contextmanager
from decimal import Decimal
from functools import cached_property
from pathlib import Path
from typing import Tuple, Union


SingularID = str
PluralID = Tuple[str, int]
MsgID = Union[SingularID, PluralID]
locale_lock = threading.Lock()
SYSTEM_LOCALES = Path('/usr/share/locale')


class Locale:
    def __init__(self, path: Path | None = None, *, language: str | None = None) -> None:
        self.path = path
        self._lang = language

    def reset_cache(self) -> None:
        """Reset all the cached values for the locale object.

        Use it if you need to reload the mo file.
        """
        path = self.path
        lang = self._lang
        vars(self).clear()
        self.path = path
        self._lang = lang

    def get(
        self,
        message: str, *,
        context: str | None = None,
        plural: str | None = None,
        n: int | None = None,
        comment: str = '',
    ) -> str:
        """Get translation for the message from the catalog (mo file).

        Args:
            message: the message to translate.
                If no translation found, the message itself will be used.
                Represented as `msgid` in PO files.
            context: kind of namespace for translations, used to distinguish
                two messages that may have a different translation depending
                on the context they are used in. Represented as `msgctxt` in PO files.
            plural: the default value if `n > 1` and no translation found.
                It also indicates for translator that the message should have
                plural translations. Represented as `msgid_plural` in PO files.
            n: the number used to pick a plural form for the translation.
            comment: not used in runtime but included in PO files.
                Use it to provide additional information for translators.
        """
        msgid_str = message
        if context is not None:
            msgid_str = f'{context}\x04{msgid_str}'
        msgid: MsgID = msgid_str
        if n is not None:
            msgid = (msgid_str, self._plural_id(n))
        translation = self._messages.get(msgid)

        if translation is not None:
            return translation
        if n is not None and n != 1:
            return plural or message
        return message

    @property
    def language(self) -> str:
        """The language of the Locale.
        """
        if self._lang:
            return self._lang
        return self._headers.get('language', '')

    def format_date(self, date: datetime.date) -> str:
        """Format date.

        You need the locale do be compiled in your OS.
        """
        format = self._get_locale_info(locale.D_FMT)
        return date.strftime(format)

    def format_time(self, time: datetime.time) -> str:
        """Format time.

        You need the locale do be compiled in your OS.
        """
        format = self._get_locale_info(locale.T_FMT)
        return time.strftime(format)

    def format_datetime(self, dt: datetime.datetime) -> str:
        """Format date and time.

        You need the locale do be compiled in your OS.
        """
        format = self._get_locale_info(locale.D_T_FMT)
        return dt.strftime(format)

    def format_month(self, n: int, *, abbreviate: bool = False) -> str:
        """Format the month number into its name.

        You need the locale do be compiled in your OS.
        """
        prefix = 'AB' if abbreviate else ''
        info_key = getattr(locale, f'{prefix}MON_{n}')
        return self._get_locale_info(info_key)

    def format_dow(self, n: int, *, abbreviate: bool = False, sunday: int = 0) -> str:
        """Format the day of week.

        You need the locale do be compiled in your OS.

        Args:
            abbreviate: Use the short version of the DoW.
            sunday: Depending on your country and if you count from 0 or 1,
                different numbers correspond to a different day of week.
                The sunday argument allows you to specify what number is Sunday for you.
                By default, it is `0` assuming that the first DoW is Sunday
                and you count days from 0 to 6.
        """
        if sunday not in {0, 1, 6, 7}:
            raise TypeError('invalid value for sunday')
        n = (n - 1) % 7 + 1
        if sunday in (0, 7):
            n += 1
            if n == 8:
                n = 1
        elif sunday == 6:
            n = (n - 6) % 7 + 1
        prefix = 'AB' if abbreviate else ''
        info_key = getattr(locale, f'{prefix}DAY_{n}')
        return self._get_locale_info(info_key)

    @cached_property
    def currency_symbol(self) -> str:
        """The symbol used to denote the currency.

        You need the locale do be compiled in your OS.
        """
        return self._get_locale_info(locale.CRNCYSTR)[1:]

    @cached_property
    def decimal_dot(self) -> str:
        """The symbol used to separate whole and fractional parts in decimal numbers.

        Also known as decimal dot, decimal comma, radix character.

        You need the locale do be compiled in your OS.
        """
        return self._get_locale_info(locale.RADIXCHAR)

    @cached_property
    def thousands_separator(self) -> str:
        """The symbol used to separate groups of digits in big numbers.

        In most of the languages, it separates groups of three digits.
        In some locales, though, digits are grouped by 4.

        https://en.wikipedia.org/wiki/Decimal_separator

        You need the locale do be compiled in your OS.
        """
        return self._get_locale_info(locale.THOUSEP)

    def format_currency(
        self,
        val: int | float | Decimal, *,
        symbol: bool = True,
        international: bool = False,
        grouping: bool = False,
    ) -> str:
        """Format a price in the locale currency.

        1. The locale currency symbol will be added.
        2. The currency symbol will be placed in the right position.
        3. The minus for negative numbers will be placed in the right position.
        4. The locale decimal separator will be used.

        Args:
            symbol: include the currency symbol.
            international: use international currency symbol.
            grouping: add thousands separator.
        """
        with self._context():
            return locale.currency(
                val,
                symbol=symbol,
                grouping=grouping,
                international=international,
            )

    def format_float(
        self,
        n: int | float | Decimal, *,
        grouping: bool = False,
        monetary: bool = False,
        precision: int | None = None,
        strip_zeros: bool | None = None,
        exp: bool = False,
    ) -> str:
        """Format float or Decimal with the correct decimal dot.

        Args:
            grouping: add thousands separator.
            monetary: use monetary thousands separator which may be different
                from a regular one for some locales.
            precision: if specified, include exactly so many digits in the decimal part.
            strip_zeros: whatever to remove trailing zeros from the end.
                If not specified, will be set to `True` only if precision
                is not specified.
            exp: if True, represent large numbers in scientific exponent notation.
        """
        fmt_suffix = 'g' if exp else 'f'
        fmt = f'%.{precision}{fmt_suffix}' if precision is not None else f'%{fmt_suffix}'
        with self._context():
            result = locale.format_string(fmt, n, grouping=grouping, monetary=monetary)
        if strip_zeros is None:
            strip_zeros = precision is None
        if strip_zeros:
            result = result.rstrip('0')
        return result

    format_decimal = format_float

    def format_int(self, n: int, grouping: bool = True) -> str:
        """Format an integer with grouping (thousands separator).

        If you need to show an integer without grouping, you don't need l10n.
        All languages display whole numbers in arabic numerals in the same way.

        https://en.wikipedia.org/wiki/Arabic_numerals
        https://en.wikipedia.org/wiki/Decimal_separator
        """
        if not grouping:
            return str(n)
        with self._context():
            return f'{n:n}'

    def parse_float(self, s: str) -> float:
        """Convert string generated by `Locale.format_float` back into float.
        """
        with self._context():
            return locale.atof(s)

    def parse_int(self, s: str) -> int:
        """Convert string generated by `Locale.format_int` back into int.
        """
        with self._context():
            return locale.atoi(s)

    def translate_country(self, country_name: str) -> str:
        """Translate country name from English to the given language.

        Follows ISO 3166-1. Requires `iso-codes` package to be installed.
        https://en.wikipedia.org/wiki/ISO_3166-1
        """
        if not self._iso_3166_1:
            return country_name
        return self._iso_3166_1.get(country_name)

    def translate_currency(self, currency_name: str) -> str:
        """Translate currency name from English to the given language.

        Follows ISO 4217. Requires `iso-codes` package to be installed.
        https://en.wikipedia.org/wiki/ISO_4217
        """
        if not self._iso_4217:
            return currency_name
        return self._iso_4217.get(currency_name)

    def translate_language(self, language_name: str) -> str:
        """Translate language name from English to the given language.

        Follows ISO 639-2 and ISO 639-3. Requires `iso-codes` package to be installed.
        https://en.wikipedia.org/wiki/ISO_639-2
        https://en.wikipedia.org/wiki/ISO_639-3
        """
        if self._iso_639_2:
            transl = self._iso_639_2.get(language_name)
            if transl != language_name:
                return transl
        if self._iso_639_3:
            transl = self._iso_639_3.get(language_name)
            if transl != language_name:
                return transl
        return language_name

    # PRIVATE

    @cached_property
    def _iso_3166_1(self) -> Locale | None:
        return self._read_system_messages('iso_3166-1.mo')

    @cached_property
    def _iso_4217(self) -> Locale | None:
        return self._read_system_messages('iso_4217.mo')

    @cached_property
    def _iso_639_2(self) -> Locale | None:
        return self._read_system_messages('iso_639-2.mo')

    @cached_property
    def _iso_639_3(self) -> Locale | None:
        return self._read_system_messages('iso_639-3.mo')

    def _read_system_messages(self, name: str) -> Locale | None:
        long_lang = locale.normalize(self.language).split('.')[0]
        short_lang = long_lang.split('_')[0].split('-')[0]
        for lang in (self.language, long_lang, short_lang):
            path = SYSTEM_LOCALES / lang / 'LC_MESSAGES' / name
            if path.exists():
                return Locale(path, language=lang)
        return None

    def _get_locale_info(self, info_key: int) -> str:
        with self._context():
            return locale.nl_langinfo(info_key)

    @contextmanager
    def _context(self):
        with locale_lock:
            old_lang, old_enc = locale.getlocale()
            locale.setlocale(locale.LC_ALL, locale.normalize(self.language))
            try:
                yield
            finally:
                locale.setlocale(locale.LC_ALL, (old_lang, old_enc))

    @cached_property
    def _messages(self) -> dict[MsgID, str]:
        if self.path is None:
            raise RuntimeError('path to mo file is not specified for the Locale')
        with self.path.open('rb') as stream:
            tr = gettext.GNUTranslations(stream)    # type: ignore[arg-type]
        self._plural_id = tr.plural                 # type: ignore
        self._headers = tr._info                    # type: ignore[attr-defined]
        return tr._catalog                          # type: ignore[attr-defined]

    @cached_property
    def _headers(self) -> dict[str, str]:
        self._messages
        return vars(self)['_headers']

    def _plural_id(self, n: int) -> int:
        return int(n != 1)

    # MAGIC METHODS

    def __repr__(self) -> str:
        return f'{type(self).__name__}(path={self.path}, language={self._lang})'

    def __eq__(self, other):
        if not isinstance(other, Locale):
            return NotImplemented
        if self.path:
            return self.path == other.path
        return self._lang == other._lang





import sys
from argparse import ArgumentParser, Namespace
from dataclasses import dataclass
from typing import TextIO


@dataclass(frozen=True)
class Command:
    args: Namespace
    stream: TextIO = sys.stdout

    @staticmethod
    def init_parser(parser: ArgumentParser) -> None:
        return None

    def run(self) -> int:
        return 0

    def print(self, text: str):
        print(text, file=self.stream)



from pathlib import Path

import polib
import pytest

from l10n import Locale
from l10n._cli import main


@pytest.fixture
def compile(project_root: Path):
    def f(*entries):
        po_file = polib.POFile(encoding='UTF-8')
        po_file.metadata['Content-Type'] = 'text/plain; charset=UTF-8'
        po_file.extend(entries)
        po_path = project_root / 'locales' / 'ru.po'
        po_path.parent.mkdir(exist_ok=True)
        po_file.save(str(po_path))
        code = main(['compile', '--path', str(project_root)])
        assert code == 0
        mo_file = project_root / 'project_test' / 'locales' / 'ru.mo'
        assert mo_file.exists()
        return Locale(mo_file)
    return f


def test_compile_simple(compile):
    e = polib.POEntry(msgid='hello world', msgstr='привет мир')
    loc: Locale = compile(e)
    assert loc.get('hello world') == 'привет мир'


def test_include_fuzzy(compile):
    e = polib.POEntry(msgid='hello world', msgstr='привет мир', flags=['fuzzy'])
    loc: Locale = compile(e)
    assert loc.get('hello world') == 'привет мир'




from types import MappingProxyType
from typing import Mapping, Type

from ._base import Command
from ._compile import Compile
from ._extract import Extract
from ._translate import Translate


COMMANDS: Mapping[str, Type[Command]]
COMMANDS = MappingProxyType(dict(
    compile=Compile,
    extract=Extract,
    translate=Translate,
))



from datetime import date, time
from pathlib import Path

import pytest

from l10n._locale import Locale


SPHINX_MO = Path('/usr/share/locale/ru/LC_MESSAGES/sphinx.mo')


@pytest.mark.skipif(not SPHINX_MO.exists(), reason='no mo file for Sphinx')
def test_read_mo():
    loc = Locale(SPHINX_MO)
    assert loc.language == 'ru'
    msgid = 'Could not import extension %s'
    assert loc.get(msgid) == 'Не могу загрузить модуль расширения %s'


@pytest.mark.parametrize('language, expected', [
    ('ru', '31.12.2021'),
    ('ru_RU', '31.12.2021'),
    ('en', '12/31/2021'),
    ('en_US', '12/31/2021'),
    ('en_UK', '31/12/21'),
    ('hu', '2021-12-31'),
    ('nl', '31-12-21'),
    ('nl_NL', '31-12-21'),
])
def test_format_date(language, expected):
    loc = Locale(language=language)
    given = date(2021, 12, 31)
    assert loc.format_date(given) == expected


@pytest.mark.parametrize('language, expected', [
    ('ru', '23:34:56'),
    ('ru_RU', '23:34:56'),
    ('en', '11:34:56 PM'),
    ('en_US', '11:34:56 PM'),
    ('en_UK', '23:34:56'),
    ('hu', '23:34:56'),
    ('nl', '23:34:56'),
    ('nl_NL', '23:34:56'),
])
def test_format_time(language, expected):
    loc = Locale(language=language)
    given = time(23, 34, 56)
    assert loc.format_time(given) == expected


@pytest.mark.parametrize('language, expected', [
    ('ru', 'ноября'),
    ('ru_RU', 'ноября'),
    ('en', 'November'),
    ('en_US', 'November'),
    ('en_UK', 'November'),
    ('hu', 'november'),
    ('nl', 'november'),
    ('nl_NL', 'november'),
])
def test_format_month(language, expected):
    loc = Locale(language=language)
    assert loc.format_month(11) == expected


@pytest.mark.parametrize('language, expected, expected_abbr', [
    ('ru', 'Понедельник', 'Пн'),
    ('ru_RU', 'Понедельник', 'Пн'),
    ('en', 'Monday', 'Mon'),
    ('en_US', 'Monday', 'Mon'),
    ('en_UK', 'Monday', 'Mon'),
    ('hu', 'hétfő', 'h'),
    ('nl', 'maandag', 'ma'),
    ('nl_NL', 'maandag', 'ma'),
])
def test_format_dow(language, expected, expected_abbr):
    loc = Locale(language=language)
    assert loc.format_dow(1) == expected
    assert loc.format_dow(1, abbreviate=True) == expected_abbr


@pytest.mark.parametrize('n, sunday, expected', [
    (0, 0, 'Sunday'),
    (1, 0, 'Monday'),
    (6, 0, 'Saturday'),

    (1, 1, 'Sunday'),
    (2, 1, 'Monday'),
    (7, 1, 'Saturday'),

    (6, 6, 'Sunday'),
    (0, 6, 'Monday'),
    (5, 6, 'Saturday'),

    (7, 7, 'Sunday'),
    (1, 7, 'Monday'),
    (6, 7, 'Saturday'),
])
def test_format_dow__sunday(n, sunday, expected):
    loc = Locale(language='en')
    assert loc.format_dow(n, sunday=sunday) == expected


@pytest.mark.parametrize('language, expected', [
    ('ru', '₽'),
    ('ru_RU', '₽'),
    ('en', '$'),
    ('en_US', '$'),
    ('en_UK', '£'),
    ('hu', 'Ft'),
    ('nl', 'EUR'),
    ('nl_NL', 'EUR'),
    ('fa_IR', 'ریال'),
])
def test_currency_symbol(language, expected):
    loc = Locale(language=language)
    assert loc.currency_symbol == expected


@pytest.mark.parametrize('language, expected', [
    ('ru', '-16,00 ₽'),
    ('ru_RU', '-16,00 ₽'),
    ('en', '-$16.00'),
    ('en_US', '-$16.00'),
    ('en_UK', '-£16.00'),
    ('hu', '-16,00 Ft'),
    ('nl', 'EUR 16,00-'),
    ('nl_NL', 'EUR 16,00-'),
    ('fa_IR', '-16 ریال'),
])
def test_format_currency(language, expected):
    loc = Locale(language=language)
    assert loc.format_currency(-16) == expected


@pytest.mark.parametrize('language, expected', [
    ('ru', '-16\u202f723'),
    ('ru_RU', '-16\u202f723'),
    ('en', '-16,723'),
    ('en_US', '-16,723'),
    ('en_UK', '-16,723'),
    ('hu', '-16.723'),
    ('nl', '-16.723'),
    ('nl_NL', '-16.723'),
    ('fa_IR', '-16,723'),
])
def test_format_parse_int(language, expected):
    loc = Locale(language=language)
    assert loc.format_int(-16723) == expected
    assert loc.parse_int(expected) == -16723


@pytest.mark.parametrize('language, expected', [
    ('ru', '-16\u202f723,34'),
    ('ru_RU', '-16\u202f723,34'),
    ('en', '-16,723.34'),
    ('en_US', '-16,723.34'),
    ('en_UK', '-16,723.34'),
    ('hu', '-16.723,34'),
    ('nl', '-16.723,34'),
    ('nl_NL', '-16.723,34'),
    ('fa_IR', '-16,723.34'),
])
def test_format_parse_float__grouping(language, expected):
    loc = Locale(language=language)
    assert loc.format_float(-16723.34, grouping=True) == expected
    assert loc.parse_float(expected) == -16723.34


@pytest.mark.parametrize('language, expected', [
    ('ru', 'Нидерланды'),
    ('ru_RU', 'Нидерланды'),
    ('en', 'Netherlands'),
    ('en_US', 'Netherlands'),
    ('en_UK', 'Netherlands'),
    ('hu', 'Hollandia'),
    ('nl', 'Nederland'),
    ('nl_NL', 'Nederland'),
    ('fa_IR', 'هلند'),
])
def test_translate_country(language, expected):
    loc = Locale(language=language)
    assert loc.translate_country('Netherlands') == expected


@pytest.mark.parametrize('language, expected', [
    ('ru', 'голландский'),
    ('ru_RU', 'голландский'),
    ('en', 'Dutch'),
    ('en_US', 'Dutch'),
    ('en_UK', 'Dutch'),
    ('hu', 'holland'),
    ('nl', 'Nederlands'),
    ('nl_NL', 'Nederlands'),
    ('fa_IR', 'هلندی'),
])
def test_translate_language(language, expected):
    loc = Locale(language=language)
    assert loc.translate_language('Dutch') == expected


@pytest.mark.parametrize('language, expected', [
    ('ru', 'Евро'),
    ('ru_RU', 'Евро'),
    ('en', 'Euro'),
    ('en_US', 'Euro'),
    ('en_UK', 'Euro'),
    ('hu', 'euró'),
    ('nl', 'euro'),
    ('nl_NL', 'euro'),
    ('fa_IR', 'Euro'),
])
def test_translate_currency(language, expected):
    loc = Locale(language=language)
    assert loc.translate_currency('Euro') == expected


PATHS = [pytest.param(p, id=p.name) for p in Path('/usr/share/locale').iterdir()]


@pytest.mark.parametrize('root', PATHS)
def test_parse_smoke(root: Path):
    msg_root = (root / 'LC_MESSAGES')
    if not msg_root.exists():
        pytest.skip()
    for path in msg_root.iterdir():
        if path.suffix != '.mo':
            continue
        loc = Locale(path)
        try:
            loc.language
        except (ValueError, IndexError):
            # for some reason, gettext can't parse some plural forms
            pass



from pathlib import Path
from textwrap import dedent

import polib
import pytest

from l10n._cli import main


@pytest.fixture
def extract(project_root: Path, source_path: Path):
    def f(source):
        source_path.write_text(dedent(source))
        code = main(['extract', '--path', str(project_root), '--lang', 'ru'])
        assert code == 0
        po_path = project_root / 'locales' / 'ru.po'
        return polib.pofile(str(po_path))
    return f


def test_extract_simple(extract):
    po_file: polib.POFile = extract("""
        from l10n import Locales
        Locales()['en'].get("hello")
    """)
    assert [e.msgid for e in po_file] == ['hello']


def test_extract_constant(extract):
    po_file: polib.POFile = extract("""
        from typing import Final
        from l10n import Locales
        msg_id: Final = "oh hi mark"
        loc = Locales()['en']
        loc.get(msg_id)
    """)
    assert [e.msgid for e in po_file] == ['oh hi mark']


def test_extract_comment(extract):
    po_file: polib.POFile = extract("""
        from l10n import Locales
        Locales()['en'].get("hello", comment="oh hi mark")
    """)
    assert [e.comment for e in po_file] == ['oh hi mark']


def test_extract_plural(extract):
    po_file: polib.POFile = extract("""
        from l10n import Locales
        Locales()['en'].get("{n} bird", plural="{n} birds", n=13)
    """)
    assert [e.msgid_plural for e in po_file] == ['{n} birds']


def test_extract_context(extract):
    po_file: polib.POFile = extract("""
        from l10n import Locales
        Locales()['en'].get("open", context="a verb")
    """)
    assert [e.msgctxt for e in po_file] == ['a verb']


def test_detect_formatting(extract):
    po_file: polib.POFile = extract("""
        from l10n import Locales
        Locales()['en'].get("hello {username}").format(username="world")
    """)
    assert [e.flags for e in po_file] == [['python-brace-format']]


def test_detect_metadata(extract):
    po_file: polib.POFile = extract("""
        from l10n import Locales
        Locales()['en'].get("open", context="a verb")
    """)
    m = po_file.metadata
    assert m['Project-Id-Version'] == 'project-test 0.0.0'
    assert m['POT-Creation-Date'] == m['PO-Revision-Date']
    assert m['Language'] == 'ru'
    assert m['Plural-Forms'].startswith('nplurals=3; plural=')


def test_update_existing(extract):
    extract("""
        from l10n import Locales
        Locales()['en'].get("hello")
    """)
    po_file: polib.POFile = extract("""
        from l10n import Locales
        Locales()['en'].get("world")
    """)
    enew, eold = [e for e in po_file]
    assert enew.msgid == 'world'
    assert eold.msgid == 'hello'
    assert eold.obsolete



from __future__ import annotations

import fnmatch
import inspect
import locale
import re
from functools import cached_property, lru_cache
from pathlib import Path
from typing import Iterator

from ._locale import Locale


class Locales:
    """Class allowing you to work with compiled (.mo) files collection.

    Args:
        path: where compiled locales are located.
        format: file name template for compiled locales.
    """
    _path: Path | None
    format: str

    def __init__(
        self, *,
        path: Path | None = None,
        format: str = '{language}.mo',
    ) -> None:
        self._path = path
        self.format = format

    def get(self, language: str) -> Locale | None:
        """Find locale for the given language.
        """
        path = self._path_to(language)
        if path.exists():
            return Locale(path, language=language)
        short_lang = language.split('_')[0].split('-')[0]
        path = self._path_to(short_lang)
        if path.exists():
            return Locale(path, language=language)
        return None

    @lru_cache(maxsize=16)
    def get_cached(self, language: str) -> Locale | None:
        """The same as get but caches the returned Locale.
        """
        return self.get(language)

    @cached_property
    def path(self) -> Path:
        if self._path is not None:
            return self._path
        return self._find_catalog()

    @cached_property
    def languages(self) -> frozenset[str]:
        """List all languages for which a locale is available in the catalog.
        """
        return frozenset(locale.language for locale in self.locales)

    @cached_property
    def locales(self) -> tuple[Locale, ...]:
        """List all locales available in the catalog.
        """
        locales = []
        for path in self.path.glob(self._pattern):
            locales.append(Locale(path))
        return tuple(locales)

    @cached_property
    def system_locale(self) -> Locale | None:
        """The Locale for the default system language.
        """
        if self.system_language is None:
            return None
        return self.get(self.system_language)

    @cached_property
    def system_language(self) -> str | None:
        """The current system language detected from env vars.
        """
        return locale.getdefaultlocale()[0]

    def reset_cache(self) -> None:
        path = self._path
        format = self.format
        vars(self).clear()
        self._path = path
        self.format = format

    # PRIVATE

    def _path_to(self, language: str) -> Path:
        parts = self.format.format(language=language).split('/')
        return self.path.joinpath(*parts)

    @cached_property
    def _rex(self) -> re.Pattern:
        return re.compile(fnmatch.translate(self._pattern))

    @cached_property
    def _pattern(self) -> str:
        return self.format.replace('{language}', '*')

    @staticmethod
    def _find_catalog() -> Path:
        for frame in inspect.stack():
            if frame.filename == __file__:
                continue
            file_path = Path(frame.filename)
            if file_path.name == 'functools.py':
                continue
            for dir_path in file_path.parents:
                catalog_path = dir_path / 'locales'
                if dir_path.exists():
                    return catalog_path
        raise FileNotFoundError('cannot find catalog')

    # MAGIC

    def __getitem__(self, language: str) -> Locale:
        """The same as Locales.get but raises KeyError if no Locale found.
        """
        locale = self.get(language)
        if locale is None:
            raise KeyError(language)
        return locale

    def __iter__(self) -> Iterator[Locale]:
        """Iterate thorugh all compiled locales.
        """
        yield from self.locales



from __future__ import annotations

import json
import subprocess
import sys
from pathlib import Path
from tempfile import NamedTemporaryFile
from textwrap import dedent
from typing import Iterator, NamedTuple

from mypy import types
from mypy.plugin import MethodContext, Plugin
from mypy.types import LiteralValue


PREFIX = '__L10N__:'
CONFIG = f"""
    [mypy]
    plugins = {__name__}
    follow_imports = skip

    [mypy-l10n.*]
    follow_imports = normal
"""


class Message(NamedTuple):
    text: str
    line: int
    column: int
    n: int | None
    context: str | None
    comment: str | None
    plural: str | None
    file_name: str

    @property
    def path(self) -> Path:
        return Path(self.file_name)


def extract_messages(project_path: Path) -> Iterator[Message]:
    config = dedent(CONFIG)
    with NamedTemporaryFile() as tmp_file:
        tmp_file.write(config.encode())
        tmp_file.flush()
        cmd = [
            sys.executable, '-m', 'mypy',
            '--no-incremental',
            '--show-traceback',
            '--check-untyped-defs',
            '--config-file', tmp_file.name,
            str(project_path),
        ]
        # result = subprocess.run(cmd)
        result = subprocess.run(cmd, stdout=subprocess.PIPE)
    for line in result.stdout.decode().splitlines():
        if not line.startswith(PREFIX):
            continue
        line = line[len(PREFIX):].strip()
        yield Message(**json.loads(line))


class LookupPlugin(Plugin):
    def get_method_hook(self, fullname: str):
        if fullname == 'l10n._locale.Locale.get':
            return self._extractor

    def _extractor(self, context: MethodContext):
        message = self._get_arg('message', context)
        if message:
            self._record(
                text=message,
                n=self._get_arg('n', context),
                context=self._get_arg('context', context),
                comment=self._get_arg('comment', context),
                plural=self._get_arg('plural', context),
                line=context.context.line,
                column=context.context.column,
                file_name=context.api.path,
            )
        return context.default_return_type

    def _get_arg(self, name: str, context: MethodContext) -> LiteralValue | None:
        index = context.callee_arg_names.index(name)
        arg_types = context.arg_types[index]
        if len(arg_types) != 1:
            return None
        arg_type = arg_types[0]
        if not isinstance(arg_type, types.Instance):
            return None
        if arg_type.last_known_value is None:
            return None
        return arg_type.last_known_value.value

    def _record(self, **data: object) -> None:
        print(PREFIX, json.dumps(data))


def plugin(version: str):
    return LookupPlugin



import os
from pathlib import Path

import pytest

from l10n import Locales


def test_system_language():
    default = os.environ.get('LC_ALL')
    os.environ['LC_ALL'] = 'nl_NL.UTF-8'
    try:
        loc = Locales()
        assert loc.system_language == 'nl_NL'
    finally:
        if default is not None:
            os.environ['LC_ALL'] = default


def test_get_locale(tmp_path: Path):
    ru_path = tmp_path / 'ru.mo'
    ru_path.write_text('')
    locales = Locales(path=tmp_path)
    assert locales['ru'].path == ru_path
    assert locales['ru_RU'].path == ru_path
    assert locales.get('ru') is not None
    assert locales.get('en') is None
    with pytest.raises(KeyError):
        locales['en']



from __future__ import annotations

import getpass
from contextlib import suppress
from dataclasses import dataclass
from functools import cached_property
from pathlib import Path
from typing import Any, Mapping


try:
    import toml
except ImportError:
    toml = None  # type: ignore[assignment]
try:
    import tomli
except ImportError:
    # https://peps.python.org/pep-0680/
    try:
        import tomllib as tomli  # type: ignore
    except ImportError:
        tomli = None   # type: ignore[assignment]


def find_project_root(start_path: Path) -> Path:
    parents = [start_path] + list(start_path.absolute().parents)
    for dir_path in parents:
        if (dir_path / 'pyproject.toml').is_file():
            return dir_path
    for dir_path in parents:
        if (dir_path / '.git').is_dir():
            return dir_path

    # find the parent directory without __init__.py that contains a directory
    # with __init__.py.
    init_detected = False
    for dir_path in parents:
        if (dir_path / '__init__.py').exists():
            init_detected = True
            continue
        if init_detected:
            return dir_path
        break

    return Path()


@dataclass(frozen=True)
class Project:
    root: Path

    @cached_property
    def name(self) -> str:
        with suppress(KeyError):
            return self._meta['tool']['l10n']['name']
        with suppress(KeyError):
            return self._meta['project']['name']
        with suppress(KeyError):
            return self._meta['tool']['flit']['metadata']['module']
        with suppress(KeyError):
            return self._meta['tool']['poetry']['name']
        return self.root.absolute().name

    @cached_property
    def version(self) -> str:
        with suppress(KeyError):
            return self._meta['tool']['l10n']['version']
        with suppress(KeyError):
            return self._meta['project']['version']
        with suppress(KeyError):
            return self._meta['tool']['poetry']['version']
        return '0.0.0'

    @cached_property
    def author_name(self) -> str:
        with suppress(KeyError):
            return self._meta['tool']['l10n']['author_name']
        with suppress(LookupError):
            return self._meta['project']['authors'][0]['name']
        return getpass.getuser()

    @cached_property
    def author_email(self) -> str:
        with suppress(KeyError):
            return self._meta['tool']['l10n']['author_email']
        with suppress(LookupError):
            return self._meta['project']['authors'][0]['email']
        return ''

    @cached_property
    def author(self) -> str:
        if self.author_email:
            return f'{self.author_name} <{self.author_email}>'
        return self.author_name

    @cached_property
    def bug_tracker(self) -> str:
        """URL or e-mail address where bugs in MsgIDs (original text) can be reported.
        """
        return self.author_email

    @cached_property
    def po_dir(self) -> str:
        """Name of the directory with po files in the project root dir.
        """
        with suppress(KeyError):
            return self._meta['tool']['l10n']['po_dir']
        return 'locales'

    @cached_property
    def po_root(self) -> Path:
        return self.root / self.po_dir

    @cached_property
    def mo_dir(self) -> str:
        """Name of the directory with mo files in the package dir.
        """
        with suppress(KeyError):
            return self._meta['tool']['l10n']['mo_dir']
        return 'locales'

    @cached_property
    def mo_root(self) -> Path:
        return self.package_path / self.mo_dir

    @cached_property
    def package_path(self) -> Path:
        """Path to the Python source code of the project.
        """
        # First try the same name as the project name
        path = self.root / self.name.replace('-', '_')
        if (path / '__init__.py').exists():
            return path

        # Also look into ./src/
        path = self.root / 'src' / self.name.replace('-', '_')
        if (path / '__init__.py').exists():
            return path

        # Try to find any path with __init__.py in it
        for path in self.root.iterdir():
            if path.name == 'tests':
                continue
            if (path / '__init__.py').exists():
                return path
        # try to find any directory with a Python file in it
        for path in self.root.iterdir():
            if path.name == 'tests':
                continue
            for _ in path.glob('*.py'):
                return path
        for _ in self.root.glob('*.py'):
            return self.root
        raise FileNotFoundError('cannot find the package')

    # PRIVATE

    @cached_property
    def _meta(self) -> Mapping[str, Any]:
        path = self.root / 'pyproject.toml'
        if not path.exists():
            return {}
        if tomli is not None:
            with path.open('rb') as stream:
                return tomli.load(stream)
        if toml is not None:   # type: ignore[unreachable]
            with path.open('rb', encoding='utf8') as stream:
                return dict(toml.load(stream))
        return {}



from __future__ import annotations

from argparse import ArgumentParser
from pathlib import Path

import polib

from .._project import Project, find_project_root
from ._base import Command


class Compile(Command):
    """Generate `.mo` files out of `.po` files.
    """
    @staticmethod
    def init_parser(parser: ArgumentParser) -> None:
        parser.add_argument(
            '--path', type=Path, default=Path(),
            help='the base directory from where to start project root lookup',
        )
        parser.add_argument(
            '--no-fuzzy', action='store_true',
            help='do not include fuzzy translations',
        )
        parser.add_argument(
            '--allow-empty', action='store_true',
            help='allow emitting `.mo` for untranslated `.po` files',
        )

    def run(self) -> int:
        project_root = find_project_root(self.args.path)
        project = Project(project_root)
        project.mo_root.mkdir(exist_ok=True)
        code = 0
        for po_path in project.po_root.iterdir():
            if po_path.suffix != '.po':
                continue
            self.print(po_path.stem)
            po_file = polib.pofile(str(po_path))

            # remove `fuzzy` flag from all entries unless `--no-fuzzy` is set.
            if not self.args.no_fuzzy:
                for entry in po_file:
                    if entry.fuzzy:
                        entry.flags.remove('fuzzy')

            # check if there is at least one translated entry
            translated = sum(e.translated() for e in po_file)
            if not self.args.allow_empty and not translated:
                self.print('  no translated strings found')
                code += 1
                continue
            self.print(f'  included: {translated}')

            mo_path = project.mo_root / f'{po_path.stem}.mo'
            po_file.save_as_mofile(str(mo_path))
        return code



from __future__ import annotations

from argparse import ArgumentParser
from collections import defaultdict
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, Iterator

import polib

import l10n

from .._extractor import Message, extract_messages
from .._plurals import GERMANIC, PLURALS
from .._project import Project, find_project_root
from ._base import Command


class Extract(Command):
    """Generate a `.po` file for a language based on messages in source code.
    """
    @staticmethod
    def init_parser(parser: ArgumentParser) -> None:
        parser.add_argument(
            '--path', type=Path, default=Path(),
            help='path to the project root directory',
        )
        parser.add_argument(
            '--lang',
            help='language to generate a PO file for',
        )
        parser.add_argument(
            '--wrap', default=90, type=int,
            help='wrap msgstr longer than that many characters',
        )
        parser.add_argument(
            '--echo', action='store_true',
            help='set msgstr to the same value as msgid',
        )
        parser.add_argument(
            '--allow-duplicates', action='store_true',
            help='do not check the file for duplicates',
        )
        now = datetime.now(timezone.utc).astimezone()
        parser.add_argument(
            '--now', default=now.isoformat(),
            help='the current time in ISO format'
        )

    def run(self) -> int:
        files: defaultdict[Path, list[polib.POEntry]]
        files = defaultdict(list)
        for msg in extract_messages(self.args.path):
            entry = self._msg_to_entry(msg)
            root = find_project_root(msg.path)
            files[root].append(entry)

        if not files:
            self.print('No entries found')
            return 1
        for root, entries in files.items():
            project = Project(root)
            project.po_root.mkdir(exist_ok=True)
            for lang in self._langs_for(project):
                self.print(lang)
                file_path = project.po_root / f'{lang}.po'

                template_file = polib.POFile(
                    wrapwidth=self.args.wrap,
                    check_for_duplicates=not self.args.allow_duplicates,
                )
                template_file.extend(entries)
                if file_path.exists():
                    target_file = polib.pofile(
                        str(file_path),
                        wrapwidth=self.args.wrap,
                        check_for_duplicates=not self.args.allow_duplicates,
                    )
                    target_file.merge(template_file)
                else:
                    target_file = template_file
                self._set_meta(project, target_file, lang)
                target_file.save(str(file_path))
                self.print(f'  extracted: {len(entries)}')
        return 0

    def _langs_for(self, project: Project) -> Iterator[str]:
        if self.args.lang:
            yield self.args.lang
            return
        found = False
        for po_file in project.po_root.iterdir():
            found = True
            yield po_file.stem
        if not found:
            yield 'en'

    def _msg_to_entry(self, msg: Message) -> polib.POEntry:
        kwargs: dict[str, Any] = {}
        if msg.comment:
            kwargs['comment'] = msg.comment
        if msg.context:
            kwargs['msgctxt'] = msg.context
        if msg.plural is not None:
            kwargs['msgid_plural'] = msg.plural
        flags = []
        if '{' in msg.text:
            flags.append('python-brace-format')
        return polib.POEntry(
            msgid=msg.text,
            msgstr=msg.text if self.args.echo else '',
            occurrences=[(msg.file_name, msg.line)],
            flags=flags,
            **kwargs,
        )

    def _set_meta(self, project: Project, po_file: polib.POFile, lang: str) -> None:
        now = datetime.fromisoformat(self.args.now).strftime('%F %H:%M%z')
        short_lang = lang.split('-')[0].split('_')[0]
        plurals = PLURALS.get(short_lang, GERMANIC)
        meta = po_file.metadata

        meta['Project-Id-Version'] = f'{project.name} {project.version}'
        meta.setdefault('Report-Msgid-Bugs-To', project.bug_tracker)
        meta.setdefault('POT-Creation-Date', now)
        meta['PO-Revision-Date'] = now
        meta.setdefault('Last-Translator', project.author)
        meta.setdefault('Language-Team', project.author)
        meta.setdefault('Language', lang)
        meta.setdefault('MIME-Version', '1.0')
        meta.setdefault('Content-Type', 'text/plain; charset=UTF-8')
        meta.setdefault('Content-Transfer-Encoding', '8bit')
        meta.setdefault('Plural-Forms', str(plurals))
        meta['Generated-By'] = f'l10n {l10n.__version__}'



from argparse import ArgumentParser
from pathlib import Path

import polib

from .._project import Project, find_project_root
from ._base import Command


class Translate(Command):
    """Translate all text without translation using Google Translate.
    """
    @staticmethod
    def init_parser(parser: ArgumentParser) -> None:
        parser.add_argument(
            '--path', type=Path, default=Path(),
            help='the base directory from where to start project root lookup',
        )
        parser.add_argument(
            '--src-lang', default='en',
            help='the language used for messages (msgid)',
        )

    def run(self) -> int:
        try:
            from googletrans import Translator
        except ImportError:
            msg = 'Please, run `python3 -m pip install googletrans==4.0.0rc1`'
            raise ImportError(msg)

        project_root = find_project_root(self.args.path)
        project = Project(project_root)
        translator = Translator()
        for po_path in project.po_root.iterdir():
            if po_path.suffix != '.po':
                continue
            self.print(po_path.stem)
            translated = 0
            po_file = polib.pofile(str(po_path))
            dest_lang = po_file.metadata.get('Language', po_path.stem)
            dest_lang = dest_lang.split('_')[0]
            dest_lang = dest_lang.split('-')[0]
            for entry in po_file:
                if entry.msgstr:
                    continue
                translation = translator.translate(
                    entry.msgid,
                    src=self.args.src_lang,
                    dest=dest_lang,
                )
                entry.msgstr = translation.text
                if 'fuzzy' not in entry.flags:
                    entry.flags.append('fuzzy')
                translated += 1
            po_file.save()
            self.print(f'  translated: {translated}')
        return 0



from pathlib import Path

import polib
import pytest

from l10n._cli import main


@pytest.fixture
def translate(project_root: Path):
    def f(*entries):
        po_file = polib.POFile(encoding='UTF-8')
        po_file.metadata['Content-Type'] = 'text/plain; charset=UTF-8'
        po_file.extend(entries)
        po_path = project_root / 'locales' / 'ru.po'
        po_path.parent.mkdir(exist_ok=True)
        po_file.save(str(po_path))
        code = main(['translate', '--path', str(project_root)])
        assert code == 0
        return polib.pofile(str(po_path))
    return f


def test_translate_simple(translate):
    e = polib.POEntry(msgid='Hello world')
    po: polib.POFile = translate(e)
    assert len(po) == 1
    e = po[0]
    assert e.msgid == 'Hello world'
    assert e.flags == ['fuzzy']
    assert e.msgstr == 'Привет, мир'
